from . import Widget
from . import Container
from .widget_tags import Tags

class Grid(Container):
    def __init__(self, parent=None, grid_x=3, grid_y=3):
        super().__init__(parent)
        self.grid_x, self.grid_y = grid_x, grid_y
        
        for y in range(self.grid_y):
            container_row = Container(self)
            container_row.style.margin = 0
            container_row.padding = 0
            container_row.layout_direction = Tags.COL
            for x in range(self.grid_x):
                grid_item = Widget(container_row)
                grid_item.padding = 2
                grid_item.margin = 0
                grid_item.style.border_thickness = 1
                grid_item.style._border_radius = 0
                container_row.add_child(grid_item)
            self.add_child(container_row)

from typing import Union

class Foo():
    def __init__(self, size=(10,10)):
        self._size = size
    
    @property
    def size(self) -> tuple:
        return self._size
    
    @size.setter
    def size(self, new_size:Union[tuple,list,int,float]) -> None:
        if type(new_size)==tuple or type(new_size)==list:
            self._size = (new_size[0], new_size[1])
        else:
            self._size = (new_size, new_size)

abc = Foo()

abc.size = 3
# is the same as
abc.size = (3,3)


import pygame
from .widget_tags import Tags
from typing import Union


class Utility():
    @staticmethod
    def Colour(colour):
        """takes an argument, and returns an RGB tuplet"""
        if type(colour) == tuple:
            return colour
        if type(colour) == list:
            return tuple(colour)
        if type(colour) == int:
            return (colour,colour,colour)
        if type(colour) == float:
            return Utility.Colour(int(colour*255))
    
    @staticmethod
    def QuadTLBR(value):
        """takes an argument, and returns a TLBR tuplet"""
        if type(value) == tuple:
            if len(value) == 4:
                return value
            if len(value) == 2:
                return (value[1], value[0],
                        value[1], value[0])
        if type(value) == list:
            if len(value) == 4:
                return tuple(value)
            if len(value) == 2:
                return (value[1], value[0],
                        value[1], value[0])
        if type(value) == int:
            return (value,value,value,value)


class Style():
    """a widget's style
    
    controls how the widget is drawn,
    and how the widget acts as a child"""

    def __init__(self, background=(200,200,200), border=(0,(0,0,0)), border_radius=5):
        self._background = background
        self._border = border
        self._border_radius = border_radius

        self._transparent_background = False
        
        self.stretch = (True,True)
        self.stretch_value = (1,1)

        self._margin = (0,0,0,0)
        self._margin_align = (Tags.CENTER,Tags.CENTER)

        self._padding = (0,0,0,0)
    
    def draw(self, surface, rect):
        if not self._transparent_background:
            pygame.draw.rect(surface=surface,
                             color=self._background,
                             rect=rect,
                             border_radius=self._border_radius)
        if self.border[0] != 0:
            border_width, border_colour = self._border
            pygame.draw.rect(surface=surface,
                             color=border_colour,
                             rect=rect,
                             width=border_width,
                             border_radius=self._border_radius)
    
    @property
    def background(self):
        return self._background
    @background.setter
    def background(self, colour):
        self._background = Utility.Colour(colour)
    def set_background(self, colour):
        self.background = colour
    
    @property
    def transparent_background(self):
        return self._transparent_background
    @transparent_background.setter
    def transparent_background(self, value:bool):
        self._transparent_background = value
    def set_transparent_background(self, value:bool):
        self.transparent_background = value
    
    @property
    def border(self):
        return self._border
    @border.setter
    def border(self, border):
        self._border = border
    def set_border(self, border):
        self.border = border
    
    @property
    def border_colour(self):
        return self._border[1]
    @border_colour.setter
    def border_colour(self, colour):
        self._border = (self._border[0],Utility.Colour(colour))
    def set_border_colour(self, colour):
        self.border_colour = colour
    
    @property
    def border_thickness(self):
        return self._border[0]
    @border_thickness.setter
    def border_thickness(self, thickness:int):
        self._border = (thickness, self._border[1])
    def set_border_thickness(self, thickness:int):
        self.border_thickness = thickness
    
    @property
    def border_radius(self):
        return self._border_radius
    @border_radius.setter
    def border_radius(self, radius:int):
        self._border_radius = radius
    def set_border_radius(self, radius:int):
        self.border_radius = radius
    
    @property
    def margin(self):
        return self._margin
    @margin.setter
    def margin(self, margin):
        self._margin = Utility.QuadTLBR(margin)
        self.ensure_positive_margin()
    
    def ensure_positive_margin(self):
        self._margin = (max(self._margin[0],0), max(self._margin[1],0),
                        max(self._margin[2],0), max(self._margin[3],0))
    
    @property
    def margin_align(self):
        return self._margin_align
    @margin_align.setter
    def margin_align(self, margin_align):
        self._margin_align = margin_align
    
    @property
    def padding(self):
        return self._padding
    @padding.setter
    def padding(self, padding):
        self._padding = Utility.QuadTLBR(padding)


"""

Widget features:

margin support:
    how much room the widget should
    have on all sides (len-4 tuple)
    (top, left, bottom, right)

    when setting margin, you can
    provide any of the following:
        * a single int, for all sides
        * two ints, for x & y margins
        * four ints, for all sides

padding support:
    how much room should be given
    around the inside of the widget's
    borders (len-4 tuple)
    (top, left, bottom, right)

    uses the same setting function
    as the margin

"""


import pygame

from typing import Union

from . import WidgetTypes
from . import Layout
from .widget_tags import Tags
from .styles import Style
from . import widget_checks

import copy
import math



class Widget():
    """basic Widget class
    
    stores:
     - name
     - size
     - location
     - children (used in containers)
     - style
    """
    
    def __init__(self, parent=None):
        self._parent = parent

        self.type = WidgetTypes.WIDGET

        self._name = 'widget' #  name
        self._classes = [] # classes
        self._size = (100,100) # size
        self._min_size = (5,5) # minimum size
        self._loc = (0,0) # location (relative to parent)

        # is_fixed, size, min_margin
        self._fixed_size = (False, (10,10), (1,1,1,1))
        # is_fixed, size
        self.fixed_min_size = (False, (0,0))

        self._children = []
        self._auto_reposition_children = False
        self._has_children = False

        self._style = Style()
        self._styles_other = {}
        self.layout = Layout()
        self.minimum_size = self.get_minimum_size()

        self._on_click = None

        self._is_focused = False
        self._is_focusable = False

        self._is_hovered = False
        self._is_hoverable = False
    
    def copy(self):
        return copy.deepcopy(self)


    # parent-linking ------------------------------------------------------- #

    @property
    def parent(self): return self._parent
    @parent.setter
    def parent(self, parent=None):
        self._parent = parent
    

    # name defining -------------------------------------------------------- #
    
    @property
    def name(self): return self._name
    @name.setter
    def name(self, name):
        self._name = name
    

    # class defining ------------------------------------------------------- #

    def add_class(self, class_name):
        self._classes.append(class_name)
    def remove_class(self, class_name):
        if self.is_class(class_name):
            self._classes.remove(class_name)
    def is_class(self, class_name):
        return class_name in self._classes
    

    # size functions ------------------------------------------------------- #

    @property
    def size(self): return self._size
    @size.setter
    def size(self, size):
        self._size = (
            max(self.minimum_size[0],size[0]),
            max(self.minimum_size[1],size[1]))
        self.reposition_children_check()
    
    def set_size(self, size):
        self.size = size
    def set_minimum_size(self, w, h):
        self._min_size = (w,h)
    
    @property
    def w(self): return self._size[0]
    @property
    def h(self): return self._size[1]
    @property
    def width(self): return self.w
    @property
    def height(self): return self.h
    @property
    def min_w(self): return self._min_size[0]
    @property
    def min_h(self): return self._min_size[1]
    @property
    def min_width(self): return self.min_w
    @property
    def min_height(self): return self.min_h

    @property
    def min_size(self): return self._min_size
    @min_size.getter
    def min_size(self, w, h):
        self._min_size = (w,h)
    
    @w.setter
    def w(self,w):
        self.size = (w,self._size[1])
    @h.setter
    def h(self,h):
        self.size = (self._size[0],h)

    @property
    def real_w(self):
        return self.w+self.margin[1]+self.margin[3]
    @property
    def real_h(self):
        return self.h+self.margin[0]+self.margin[2]
    
    @property
    def fixed_size(self): return self._fixed_size
    def set_fixed_size(self, fixed=True, size=(5,5), min_margin=(1,1,1,1)):
        """set a widget's fixed size, useful for in fluid widgets
        when you want to keep them a constant size"""
        self._fixed_size = (fixed, size, min_margin)
        self.size = size
    
    @property
    def fluid_size(self): return self._style.stretch_value
    @fluid_size.setter
    def fluid_size(self, stretch_value):
        self.set_fluid_size(*stretch_value)

    def set_fluid_size(self,x=1,y=1):
        """sets how a widget should act in a parent widget
        
        x & y should both be either floats or ints,
        and x is the "weight" of this widget if it's in a row,
        and y is the "weight" of this widget if it's in a column.
        
        weights of a widget basically allow it to take up
        more space in organisation"""
        sx = bool(x)
        sy = bool(y)
        self._style.stretch = (sx,sy)
        self._style.stretch_value = (x if sx else 1,
                                     y if sy else 1)
        if self.parent:
            self.parent.reposition_children_check()
    

    # position functions --------------------------------------------------- #

    @property
    def loc(self): return self._loc
    @loc.setter
    def loc(self, loc):
        self._loc = loc
    
    @property
    def position(self): return self._loc
    @position.setter
    def position(self, position):
        self._loc = position
    
    @property
    def x(self): return self._loc[0]
    @property
    def y(self): return self._loc[1]

    @x.setter
    def x(self,x):
        self.loc = (x,self._loc[1])
    @y.setter
    def y(self,y):
        self.loc = (self._loc[0],y)
    
    @property
    def scrn_x(self):
        if not self._parent: return self.x
        return self.x+self._parent.scrn_x
    @property
    def scrn_y(self):
        if not self._parent: return self.y
        return self.y+self._parent.scrn_y

    @property
    def center(self):
        return (int(self.scrn_x + self.real_w/2 - self.margin[3]),
                int(self.scrn_y + self.real_h/2 - self.margin[2]))
    

    # padding & margins ---------------------------------------------------- #

    @property
    def padding(self):
        return self.style.padding
    @padding.setter
    def padding(self, padding):
        self.style.padding = padding
        self.reposition_children_check()
    
    @property
    def margin(self):
        return self.style.margin
    @margin.setter
    def margin(self, margin):
        self.style.margin = margin
        self.reposition_children_check()


    # borders & rects ------------------------------------------------------ #

    @property
    def rect(self):
        """the rect attribute of the widget,
        in screen coordinates"""
        margin = self.margin
        return (self.scrn_x+margin[1], self.scrn_y+margin[0],
                self.w-margin[3]-margin[1], self.h-margin[2]-margin[0])
    
    @property
    def borders(self):
        mt, ml, mb, mr = self.margin
        scrn_x, scrn_y = self.scrn_x+ml, self.scrn_y+mt
        return scrn_x, scrn_y, scrn_x+self.w-ml-mr, scrn_y+self.h-mt-mb
    
    def loc_within_borders(self, loc):
        x,y,mx,my = self.borders
        lx,ly = loc
        return lx>x and lx<mx and ly>y and ly<my
    

    # layouts & styles ----------------------------------------------------- #
    
    @property
    def layout(self):
        return self._layout
    @layout.setter
    def layout(self, layout):
        self._layout = layout
        self.reposition_children_check()
    
    @property
    def layout_direction(self):
        return self._layout.direction
    @layout_direction.setter
    def layout_direction(self, direction):
        self._layout.direction = direction
        self.reposition_children_check()

    @property
    def style(self): return self._style
    @style.setter
    def style(self, style:Style):
        self._style = style
        if self.parent:
            self.parent.reposition_children_check()
    
    def style_copy(self):
        return copy.deepcopy(self._style)
    
    def style_other_add(self, function, style):
        self._styles_other[function] = style
    
    def style_other_remove(self, function):
        if function in self._styles_other:
            self._styles_other.pop(function)
    
    def get_style_for(self, check=widget_checks.Checks.is_focused) -> Union[Style,None]:
        if type(check) == str:
            return self._styles_other.get(widget_checks.Shortnames[check],
                                          None)
        else:
            return self._styles_other.get(check,None)
    

    # drawing/rendering ---------------------------------------------------- #

    def ovr_draw(self, surface):
        """function intended to be overwritten"""
    
    def draw(self, surface):
        """draw the widget onto the surface"""
        draw_in_style = self._style
        for function, style in self._styles_other.items():
            if function(self):
                draw_in_style = style
                break
        draw_in_style.draw(surface, self.rect)
        self.ovr_draw(surface)
        for child in self._children:
            child.draw(surface)

    
    # event-linking functionality ------------------------------------------ #

    @property
    def on_click(self): return self._on_click
    @on_click.setter
    def on_click(self, function):
        """set a widget's on_click attribute to a function"""
        if function:
            self.type = WidgetTypes.BUTTON
        self._on_click = function
    
    @property
    def is_focused(self):
        return self._is_focused
    @is_focused.setter
    def is_focused(self, value):
        self._is_focused = value
    @property
    def is_focusable(self):
        return self._is_focusable
    def set_focusable(self, focusable, style_on_focus=None):
        self._is_focusable = focusable
        check = widget_checks.Checks.is_focused
        if focusable:
            added_style = style_on_focus or self.style_copy()
            self.style_other_add(check, added_style)
        else:
            self.style_other_remove(check)
    
    @property
    def is_hovered(self):
        return self._is_hovered
    @is_hovered.setter
    def is_hovered(self, value):
        self._is_hovered = value
    @property
    def is_hoverable(self):
        return self._is_hoverable
    def set_hoverable(self, hoverable, style_on_focus=None):
        self._is_hoverable = hoverable
        check = widget_checks.Checks.is_hovered
        if hoverable:
            added_style = style_on_focus or self.style_copy()
            self.style_other_add(check, added_style)
        else:
            self.style_other_remove(check)
    

    # minimum-size calculations for scaling -------------------------------- #

    def ovr_add_minimum_size(self):
        """function intended to be overwritten"""
        return (0,0)

    def get_minimum_size(self, debug=False, indent=0):
        """returns the minimum size of the widget,
        taking all children widgets into account"""

        # set the min_size to the widget's padding & margin
        min_size = [self.min_w,
                    self.min_h]

        fixed_size = self.fixed_size
        if fixed_size[0]:
            size, margins = fixed_size[1], fixed_size[2]
            min_size = [size[0] + margins[1] + margins[3],
                        size[1] + margins[0] + margins[2]]
            return min_size

        if self.fixed_min_size[0]:
            return self.fixed_min_size[1]

        min_size[0] += self.padding[1]
        min_size[0] += self.padding[3]
        min_size[0] += self.margin[1]
        min_size[0] += self.margin[3]
        
        min_size[1] += self.padding[0]
        min_size[1] += self.padding[2]
        min_size[1] += self.margin[2]
        min_size[1] += self.margin[0]

        added_minimum_size = self.ovr_add_minimum_size()
        min_size[0] += added_minimum_size[0]
        min_size[1] += added_minimum_size[1]

        return tuple(min_size)
    

    # fluid margins -------------------------------------------------------- #

    def adjust_margins_to_area(self, area_x, area_y):
        """adjust a widget's margins to make it
        a set width and height within an area"""
        _, target_size, min_margin = self.fixed_size

        self.size = (area_x, area_y)

        differenceX = area_x - target_size[0]
        differenceY = area_y - target_size[1]

        style = self.style

        if style.margin_align[0] == Tags.CENTER:
            margin_L = math.floor(differenceX/2)
            margin_R = math.ceil(differenceX/2)
        elif style.margin_align[0] == Tags.LEFT:
            margin_L = min_margin[1]
            margin_R = differenceX-min_margin[1]
        elif style.margin_align[0] == Tags.RIGHT:
            margin_L = differenceX-min_margin[3]
            margin_R = min_margin[3]
        
        if style.margin_align[1] == Tags.CENTER:
            margin_T = math.floor(differenceY/2)
            margin_B = math.ceil(differenceY/2)
        elif style.margin_align[1] == Tags.TOP:
            margin_T = min_margin[0]
            margin_B = differenceY-min_margin[0]
        elif style.margin_align[1] == Tags.BOT:
            margin_T = differenceY-min_margin[2]
            margin_B = min_margin[2]
        
        self.margin = (margin_T,margin_L,
                       margin_B,margin_R)
    

    # child repositioning -------------------------------------------------- #

    def reposition_children_check(self):
        """run self.reposition_children()
        only if auto-reposition is enabled"""
        if self._auto_reposition_children:
            self.reposition_children()
    
    def reposition_children(self):
        """reposition own children based on layout
        - function indended to be overwritten"""


class Checks():
    """a few "check" functions for widgets
    
    is_focused, is_hovered"""

    @staticmethod
    def is_focused(widget):
        return widget.is_focused
    
    @staticmethod
    def is_hovered(widget):
        return widget.is_hovered

Shortnames = {
    'focused': Checks.is_focused,
    'hovered': Checks.is_hovered
}


import pygame
from . import widget
from . import WidgetTypes
from . import Layout
from .widget_tags import Tags

import math

class Container(widget.Widget):
    """basic Container class

    built off Widget class"""
    
    def __init__(self, parent=None):
        super().__init__(parent)
        self.type = WidgetTypes.CONTAINER
        self._has_children = True
    

    # auto-child-repositioning settings ------------------------------------ #
    
    @property
    def auto_reposition_children(self):
        return self._auto_reposition_children
    @auto_reposition_children.setter
    def auto_reposition_children(self, value):
        self._auto_reposition_children = value
    

    # child adding --------------------------------------------------------- #

    def add_child(self, child):
        """add one child to the widget's children"""
        self.add_children([child],)

    def add_children(self, children:list):
        """add a list of children to the widget's children"""
        self._children.extend(children)
        for child in children:
            child.parent = self
        self.reposition_children()

    
    # child removing ------------------------------------------------------- #

    def remove_child(self, child):
        """remove child from the widget's children
        
        returns True/False depending on whether the child
        was in the widget's children list"""
        if not child in self._children:
            return False
        self._children.remove(child)
        child.parent = None
        return True
    
    def remove_children(self, children):
        """remove a list of children from the widget's children"""
        for child in children:
            self.remove_child(child)
    

    # child getting/filtering and deep searching --------------------------- #

    def child_at(self, index):
        """get a child at an index"""
        return self._children[index]
    
    def all_children(self):
        """return a list of all children (deep search)"""
        children = []
        for child in self._children:
            children.append(child)
            if len(child._children)>0:
                children.extend(child.all_children())
        return children
    
    def filter_children(self, filter_function):
        """return a list of all children that match the filter"""
        matches = filter(filter_function,
                         [child for child in self.all_children()])
        return matches
    
    def filter_children_top(self, filter_function):
        """returns a list of all children of this widget that
        match the provided filter"""
        matches = [
            child for child in self._children if filter_function(child)]
        return matches
    

    # run functions on children -------------------------------------------- #

    def run_functions_on_children(self, function, all_children=True):
        """run a given function on this widget's children
        all_children specifies whether to run the function
        for every child under this widget (deep search),
        or just to run it for each top-level child
        
        returns the amount of children that the function
        was run on"""
        if all_children:
            run_function_on = self.all_children()
        else:
            run_function_on = self._children
        for child in run_function_on:
            function(child)
        return len(run_function_on)
    

    # minimum size check --------------------------------------------------- #
    
    def ovr_add_minimum_size(self):
        layout = self.layout
        min_size = [0,0]

        # if this widget has children, add up their sizes
        if len(self._children) > 0:

            if layout.size == Tags.FIT:
                if layout.direction == Tags.ROW:
                    i0, i1 = 0, 1
                elif layout.direction == Tags.COL:
                    i0, i1 = 1, 0
                
                # flip the axis being calculated
                # depending on whether the child widgets
                # are on a row or column

                largest = 0
                largest2 = 0
                child_sizes = []

                for child in self._children:
                    child_min_size = child.get_minimum_size()
                    if child.style.stretch[i0]:
                        weight = child.style.stretch_value[i0]
                        value = child_min_size[i0]/weight
                        child_sizes.append((value,weight),)
                    else:
                        min_size[i0] += (child.w if i0==0 else child.h)
                    largest = max(largest, child_min_size[i1])
                    largest2 = max(largest2, child_min_size[i0])

                child_sizes.sort(key=lambda i: i[0], reverse=True)

                if len(child_sizes) > 0:
                    largest_v = child_sizes[0][0]
                    for _,w in child_sizes:
                        min_size[i0] += largest_v*w
                
                else:
                    min_size[i0] += largest2

                min_size[i1] += largest
        
        return min_size
    

    # fluid area calculations ---------------------------------------------- #

    def stretch_to_max(self, children, axis=Tags.X):
        """return the maximum stretch value (the sum of
        all child widget's stretch value in a given
        axis"""
        return sum([child.style.stretch_value[axis] for child in children])
    
    def get_fluid_area(self, axis=Tags.X):
        if axis == Tags.X:
            cover_area = self.w
            cover_area -= sum([child.w for child in self.filter_children_top(
                lambda c: not c.style.stretch[0])])
            cover_area -= self.style.padding[1]
            cover_area -= self.style.padding[3]
            cover_area -= self.style.margin[1]
            cover_area -= self.style.margin[3]
            if self.name == 'debug':
                print(self.size)
                print("fluid area X:", cover_area)
        elif axis == Tags.Y:
            cover_area = self.h
            cover_area -= sum([child.h for child in self.filter_children_top(
                lambda c: not c.style.stretch[1])])
            cover_area -= self.style.padding[0]
            cover_area -= self.style.padding[2]
            cover_area -= self.style.margin[0]
            cover_area -= self.style.margin[2]
        return cover_area
    
    def align_fluid_children(self):
        """automatically adjust child widget's
        fluid values depending on their minimum
        sizes in relation to each other"""
        # get fluid area depending on whether this
        # widget is aligned as ROW or COL
        layout = self.layout
        if layout.direction == Tags.ROW:
            # calculate how much space is available
            # for fluid widgets to use up
            fluid_area = self.get_fluid_area(Tags.X)
            # get a list of all children which are fluid
            # in this axis
            fluid_children = self.filter_children_top(
                lambda c: c.style.stretch[0])
            # iterate through all the fluid children,
            # and work out their ratio for space that
            # they should take up
            for child in fluid_children:
                min_w = child.get_minimum_size()[0]
                ratio = min_w/fluid_area
                child.fluid_size = (ratio, child.fluid_size[1])
        elif layout.direction == Tags.COL:
            # calculate how much space is available
            # for fluid widgets to use up
            fluid_area = self.get_fluid_area(Tags.Y)
            # get a list of all children which are fluid
            # in this axis
            fluid_children = self.filter_children_top(
                lambda c: c.style.stretch[1])
            # iterate through all the fluid children,
            # and work out their ratio for space that
            # they should take up
            for child in fluid_children:
                min_h = child.get_minimum_size()[1]
                ratio = min_h/fluid_area
                child.fluid_size = (child.fluid_size[0], ratio)
    
    def reposition_children(self):
        """reposition own children based on layout"""

        self.minimum_size = self.get_minimum_size()

        # if this widget has no children, don't do anything
        # otherwise we'll run into x/0 division errors
        if len(self._children) == 0: return

        layout = self.layout

        if layout.direction == Tags.ROW:
            if layout.size == Tags.FIT:
                # all child widgets should be even width

                # get a list of children which are allowed to stretch on the x axis
                fluid_width_children = self.filter_children_top(lambda c: c.style.stretch[0])
                
                # calculate the area of space which is given to stretchable children
                stretch_width = self.get_fluid_area(Tags.X)

                # calculate the sum of all the stretch values of fluid children
                # this is useful, because some children will need to take up 2x the room
                # that another fluid child takes up - so this allows it to take up that amount
                stretch_to_max = self.stretch_to_max(fluid_width_children, Tags.X)

                # calculate the width of each stretchable child according to the maximum area
                width_individual = 1
                if stretch_to_max > 0:
                    width_individual = stretch_width/stretch_to_max

                # calculate the individual height of each child widget
                height_individual = self.h
                height_individual -= self.style.padding[0] # T padding (top)
                height_individual -= self.style.padding[2] # B padding (bottom)
                height_individual -= self.style.margin[0] # T margin (top)
                height_individual -= self.style.margin[2] # B margin (bottom)

                # track the x position for each child widget,
                # so we can move it according to each child's width
                x_position = self.style.padding[1] + self.style.margin[1]

                # change the widths of all the fluid-width children
                for child in self._children:
                    # set the child's width to the calculated value, multiplied by its
                    # stretch_value (x) - which is it's multiplier to take up more room

                    widget_width = child.w
                    if child in fluid_width_children:
                        widget_width = width_individual*child.style.stretch_value[0]

                        if not child.fixed_size[0]:
                            child.w = int(widget_width)
                        else:
                            available_space = int(widget_width)
                            child.adjust_margins_to_area(available_space, height_individual)

                    child.x = x_position
                    x_position += widget_width
                
                for child in self._children:
                    child.h = height_individual
                    if child.fixed_size[0]:
                        child.adjust_margins_to_area(child.w, height_individual)
                    # set the child's y position to this widget's padding
                    # basically, push it down a bit to give it room from
                    # the top border
                    child.y = self.style.padding[0] + self.style.margin[0]
        
        elif layout.direction == Tags.COL:
            if layout.size == Tags.FIT:
                # all child widgets should be even height

                # get a list of children which are allowed to stretch on the y axis
                fluid_height_children = self.filter_children_top(lambda c: c.style.stretch[1])
                
                # calculate the area of space which is given to stretchable children
                stretch_height = self.get_fluid_area(Tags.Y)

                # calculate the sum of all the stretch values of fluid children
                # this is useful, because some children will need to take up 2x the room
                # that another fluid child takes up - so this allows it to take up that amount
                stretch_to_max = self.stretch_to_max(fluid_height_children, Tags.Y)

                # calculate the height of each stretchable child according to the maximum area
                height_individual = 1
                if stretch_to_max > 0:
                    height_individual = stretch_height/stretch_to_max

                # calculate the individual width of each child widget
                width_individual = self.w
                width_individual -= self.style.padding[1] # L padding (left)
                width_individual -= self.style.padding[3] # R padding (right)
                width_individual -= self.style.margin[1] # L margin (left)
                width_individual -= self.style.margin[3] # R margin (right)

                # track the y position for each child widget,
                # so we can move it according to each child's height
                y_position = self.style.padding[0] + self.style.margin[0]

                # change the widths of all the fluid-height children
                for child in self._children:
                    # set the child's height to the calculated value, multiplied by its
                    # stretch_value (y) - which is it's multiplier to take up more room

                    widget_height = child.h
                    if child in fluid_height_children:
                        widget_height = height_individual*child.style.stretch_value[1]
                        
                        if not child.fixed_size[0]:
                            child.h = int(widget_height)
                        else:
                            available_space = int(widget_height)
                            child.adjust_margins_to_area(width_individual, available_space)
                    
                    child.y = y_position
                    y_position += widget_height
                
                for child in self._children:
                    child.w = width_individual
                    if child.fixed_size[0]:
                        child.adjust_margins_to_area(width_individual, child.h)
                    # set the child's x position to this widget's padding
                    # basically, push it right a bit to give it room from
                    # the left border
                    child.x = self.style.padding[1] + self.style.margin[1]

        for child in self._children:
            if child._has_children:
                child.reposition_children()


__version__ = '0.1.1'

from .layout import Layout

from .widget_types import WidgetTypes
from .styles import Style

from . import widget_checks
from .widget import (pygame, Widget)

from .container import Container
from .grid import Grid

from .window import Window

from .image import Image
from .label import Label


from .widget_tags import Tags

class Layout():
    """a Layout,

    used to set a widget's child
    positions and sizes
    
    attributes:
     - direction : ROW, COL, ~GRID~
     - size : FIT, TOP, BOT
    (all attributes of widget_tags.Tags)
     
    size: FIT allows the children
    widgets to squash and stretch
    to fill the given space
    
    size: TOP snaps children
    widgets to the top, and orders
    them downwards
    
    size: BOT is basically the same
    as TOP, except they're snapped
    to the bottom instead"""

    def __init__(self, direction=Tags.ROW, size=Tags.FIT):
        self.direction = direction
        self.size = size


import pygame
from . import container

class Window(container.Container):
    def __init__(self):
        super().__init__()
    
    def process_events(self, pygame_events):
        """processes pygame events, like mouse clicks if needed"""
        mouse_position = pygame.mouse.get_pos()

        # set the is_hovered attribute for hoverable child widgets
        hoverable_children = list(self.filter_children(lambda child: child.is_hoverable))
        hovered_children = list(filter(
            lambda child: child.loc_within_borders(mouse_position),
            hoverable_children))
        for child in hoverable_children:
            child.is_hovered = child in hovered_children

        clicks = []
        for event in pygame_events:
            if event.type == pygame.MOUSEBUTTONUP:
                if event.button == 1:
                    # LMB clicked and released
                    clicks.append(1)
        
        if len(clicks)>0:
            clickable_widgets = self.filter_children(
                lambda child: child.on_click)
            if 1 in clicks:
                # LMB was clicked
                # for each clickable widget, check
                # if the mouse position is over it
                for widget in clickable_widgets:
                    if widget.loc_within_borders(mouse_position):
                        widget.on_click()


import pygame
from . import WidgetTypes
from . import Widget

class Image(Widget):
    def __init__(self, parent=None, image:pygame.Surface=None):
        super().__init__(parent)
        self._image=  image
        self.set_image(image)
        self.type = WidgetTypes.IMAGE
    
    @property
    def image(self):
        return self._image
    @image.setter
    def image(self, image):
        self.set_image(image)
    
    def set_image(self, image):
        """set the widget's image"""
        self._image = image
        self.image_size = self._image.get_size()
        self.image_size_half = (int(self.image_size[0]/2),
                                int(self.image_size[1]/2))
        self.set_minimum_size(*self._image.get_size())
    
    def ovr_draw(self, surface):
        """draw the image widget"""
        center = self.center
        image_pos = (center[0]-self.image_size_half[0],
                     center[1]-self.image_size_half[1])
        surface.blit(self.image, image_pos)
        # pygame.draw.circle(surface, (100,255,100), center, 7, 2)


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pygame-ui-mb",
    version="0.1.1",
    author="MACHINE_BUILDER",
    description="A package for fluid uis in PyGame",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Machine-builder/pygame-ui",
    project_urls={
        "Bug Tracker": "https://github.com/Machine-builder/pygame-ui/issues"
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent"
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6"
)


class Tags:
    """tags are basically references for numbers
    that represent different things. It's easier
    than remembering what each different number
    corresponds to
    
    available tags:
    - ROW
    - COL
    - GRID
    
    - FIT
    - TOP
    - BOT
    
    - LEFT
    - CENTER
    - RIGHT
    
    - X
    - Y"""

    ROW = 0
    COL = 1
    GRID = 2

    FIT = 0
    TOP = 1
    BOT = 2

    # used for alignment of fluid margins
    LEFT = 0
    CENTER = 1
    RIGHT = 2

    # used for buttons
    TEXT = 0
    IMAGE = 1
    BLANK = 2

    # used for axis calculations
    X = 0
    Y = 1


import pygame

from . import WidgetTypes
from . import Image

class Label(Image):
    """basic label

    can display text"""

    def __init__(self, parent=None, font:pygame.font.Font=None, text=''):
        self.font = font
        self.set_text(text)
        super().__init__(parent, self.image)
        self.type = WidgetTypes.LABEL
    
    def set_text(self, text):
        self.image = self.font.render(text,True,(0,0,0))